"""
    Dummy conftest.py for carg_io.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""

import pytest
from carg_io.abstracts import Parameter, ParameterSet, units, NaN

class Box(ParameterSet):
    Length:Parameter = 1 * units.meter
    Width:Parameter = 1* units.meter
    Height:Parameter = 1* units.meter


import pytest
from carg_io.abstracts import Parameter, ParameterSet, units, NaN
from carg_io.spaces import Space
import numpy as np

__author__ = "eelco van Vliet"
__copyright__ = "eelco van Vliet"
__license__ = "MIT"


def test_space():
    class Box(ParameterSet):
        Length:Parameter = 1 * units.meter
        Width:Parameter = 1* units.meter
        Height:Parameter = 1* units.meter

        @property
        def Volume(self) -> Parameter:
            l = self.Length['m']
            w = self.Width['m']
            h = self.Height['m']
            return Parameter('Volume', l*w*h*units.meter**3)

    
    space = Space(Box)
    space.expand(Box.Length, 'm', np.linspace(1,10,10))
    space.expand(Box.Width, 'm', np.linspace(1,10,10))
    space.expand(Box.Height, 'm', np.linspace(1,10,10))

    space.add_criteria("Volume", 'm**3', lambda v: v < 10*10*9)

    # There should be four cases that will not pass this criteria:
    # everything 10 m, or any combination of two parameters being 10 m.
    # = 4 cases


    assert len(space) == 10**3-4
    





import pytest
from carg_io.abstracts import Parameter, ParameterSet, units, NaN
from conftest import Box

__author__ = "eelco van Vliet"
__copyright__ = "eelco van Vliet"
__license__ = "MIT"


def test_parameter1():
    
    height = Parameter('height', 8*units.meter)
    assert height.is_default
    assert height.name == 'height'
    assert height['mm'] == 8000


def test_parameter_set():

    box = Box()
    assert box.Length['km'] == 1e-3
    
    
def test_parameter_errors():
    box = Box()
    
    with pytest.raises(AssertionError):
        assert box.Length == 1

    
def test_hash():
    box = Box()
    box.Height['m'] = 99
    identical_box = Box()
    
    identical_box.Height['mm'] = 99_000
    hash1 = hash(box)
    hash2 = hash(identical_box)
    assert hash1 == hash2 # the hash checks if the instance values are equal
    assert box != identical_box # the instances themselves are not equal

def test_pickle():
    box = Box()
    file = box.to_pickle()






from carg_io.abstracts import Parameter, ParameterSet, units, NaN
from carg_io.spaces import Space
from carg_io.postprocessing import Analyze
import numpy as np


class IBeam(ParameterSet):
    Height:Parameter = 0.3 * units.m
    Width:Parameter = 0.2 * units.m
    Length:Parameter = 10 * units.m
    ThicknessWeb:Parameter = 10 * units.mm
    ThicknessFlange:Parameter = 10 * units.mm
    E:Parameter = 210 * units.GPa
    Density:Parameter = 7850 * units.kg/units.m**3

class IBeamOutput(ParameterSet):
    Volume:Parameter = NaN * units.m**3
    Mass: Parameter = NaN * units.kg
    CrossSectionArea = NaN * units.m**2
    SurfaceArea = NaN * units.m**2
    Ixx: Parameter = NaN * units.m**4
    MaxDeflextionPerTonne = NaN * units.mm
    

def calculate(ibeam:IBeam) -> IBeamOutput:
    out = IBeamOutput()

    tw = ibeam.ThicknessWeb['m']
    tf = ibeam.ThicknessFlange['m']
    w = ibeam.Width['m']
    h = ibeam.Height['m']
    l = ibeam.Length['m']

    Aw = tw * h 
    Af = tf * h
    If = 1/12 * w * tf**3
    If += (Af * (h/2)**2)
    Iw = 1/12 * tw * h**3
    out.Ixx['m**4'] = I = Iw + 2*If

    out.CrossSectionArea['m**2'] = Atot = Aw + 2 * Af
    out.Volume['m**3'] = V = Atot * l
    out.Mass['kg'] = V * ibeam.Density['kg/m**3']
    out.MaxDeflextionPerTonne['m'] = 1000 * 9.81 * l / (48 * ibeam.E['Pa'] * I)
    out.SurfaceArea['m**2'] = w*l*4 + h*l*2 + Atot*2
    return out


space = Space(IBeam)

space.expand(IBeam.Length, 'm', np.linspace(5, 15, 4))
space.expand(IBeam.Height, 'm', np.linspace(0.1, 0.5, 4))
space.expand(IBeam.ThicknessFlange, 'mm', np.linspace(8, 30, 4))
space.expand(IBeam.ThicknessWeb, 'mm', np.linspace(8, 30, 4))

results = []
for i in space.construct():
    o = calculate(i)
    results.append((i,o))

analysis = Analyze(results)
analysis.get_double_scatter(show=True)







from __future__ import annotations
import sys
from abc import ABC, ABCMeta, abstractmethod
from typing import Set, Callable, List, Dict, Tuple, Generator
import pandas as pd
import numpy as np
from copy import deepcopy
import tkinter as tk
import pint
from decimal import Decimal
from pathlib import Path
import pickle
units = ureg = pint.UnitRegistry()
NaN = np.nan

class MetaParameterSet(type):
    """
    This type ensures that init ParameterSet creates NEW instances of all
    input parameters.
    """
    def __new__(cls, clsname, bases, clsdict):
        
        if clsname=='ParameterSet': # Ignore the super
            return super().__new__(cls, clsname, bases, clsdict)
        
        # Allocate parameter container
        parameters = []

        # Extend with parameters from super (recursive by definition)
        for base in bases:
            if hasattr(base, '_parameters'):
                parameters.extend(base._parameters)

        # Extend with own parameters
        
        for name, val in clsdict.items():
            if not name.startswith('_'):
                if isinstance(val, property):
                    continue # @property may be used to create dependent Parameter
                clsdict[name] = Parameter(name, val) # Value->Parameter substitution
                parameters.append(name)
                
        clsdict['_parameters'] = parameters

        # Store ParameterSet name # NOTE: may be replaced with __qualname__
        clsdict['name'] = clsname

        return super().__new__(cls, clsname, bases, clsdict)

class Parameter():
    """The Parameter class offers a lot of functionality aound a single value.
    The value may be an input or an output value.
    It supports both input and output parameters.
    It uses Pint to support units and their conversion.

    - Getting and setting values using different units
    - Store a default value, and remember if it was changed
    - `Tkinter` `entry` representation
    
    ```python
    from carg_io import Parameter, units
    height = Parameter('height', 1.93 * units.meter)
    ```
    """

    
    def __init__(self, name, value: pint.Quantity|float|int):
        """Create a Parameter instance with value.

        value: Either a pint.Quantity, int or float. If int or float, the value will be
                converted into a dimensionless pint.Quantity.

        """
        if not isinstance(value, pint.Quantity):
            try:
                float(value)
                value = pint.Quantity(value) # Make dimensionless
            except Exception as err:
                raise TypeError(f'Parameter expect a number-like value, got {value}')
        
        self.name = name
        self._value = value
        self._unit_default = value.u # default unit
        self.is_default = True

    @property
    def normalized_value(self) -> float|int:
        """The normalized value of a `Parameter` is the value represented in the default unit.
        Any resulting zero decimals (as result of a unit conversion) are dropper (normalized).
        """#doctag[normalized_value]
        value = self._value.m_as(self._unit_default)
        nvalue = Decimal(value).normalize()
        return nvalue


    def __getitem__(self, unit:str|pint.Unit) -> pint.Quantity:
        """Get the value in the requested unit. For dimensionless units, use
        Parameter[None] or Parameter[:].
        """
        if unit == None or isinstance(unit, slice):
            return self._value.m
        return self._value.m_as(unit)

    def __setitem__(self, unit:str|pint.Unit, value:float|int):
        """Set the value with the requested unit. For dimensionless units, use
        Parameter[None] or Parameter[:].
        """
        # Check dimensionality
        try:
            self._value.m_as(unit)
        except pint.errors.DimensionalityError:
            raise ValueError(f"Trying to set {str(unit)} on parameter {self.name} that has base unit {str(self._unit_default)}")

        self._value = ureg.Quantity(value, unit)
        self.is_default = False


    def _to_tk(self, root) -> Tuple[tk.Label, tk.Entry, tk.Labe ]:
        """Return tkinter representation consiting of 2 labels (name and unit)
        and an entry field.
        
        Input:
            tkinter container entity

        Returns:
            Label(name), Entry(value), Label(unit)

        """

        name_label = tk.Label(root, text=self.name, anchor="w", width=22)

        entry = tk.Entry(root, bd = 1, width=12)
        entry.insert(0, self._value.m)
        unit = str(self._value.u)
        if unit == 'dimensionless':
            unit = '-'
        unit_label = tk.Label(root, text=unit, anchor="w", width=17)
        return name_label, entry, unit_label

    def __repr__(self):
        return f"Parameter(name='{self.name}'', value={self._value})"


class SettingAttributeNotAllowed(Exception):
    pass


class ParameterSet(metaclass=MetaParameterSet):
    """The ParameterSet class represents a collection of Parameter instances.
    
    """
    
    name:str
    _parameters:list
    
    LOCK_SETATTR = False

    def __init__(self):
        
        # Make copies of the default parameter instances
        for parm_name in self._parameters:
            parm = getattr(self, parm_name)
            setattr(self, parm_name, deepcopy(parm))
        self.LOCK_SETATTR = True
    
    @staticmethod
    def from_pickle(file:Path) -> ParameterSet:
        with open(file, 'rb') as f:
            content:ParameterSet = pickle.load(f)
        return content

    @staticmethod
    def from_serial_pickle(data) -> ParameterSet:
        content:ParameterSet = pickle.loads(data)
        return content

    def to_dataframe(self, include_set_name=False, name_include_unit=False) -> pd.DataFrame:
        """Return the dictionary representation of the instance.
        
        Parameters:
            include_set_name:bool
                If true, prepend the ParameterSet name to its Parameter names.
                This is useful when concatenating multiple ParameterSets with 
                overlapping Parameter names. Default is False.

        """
        stack = []
        for parameter in self:
            p:pint.Quantity = parameter._value
            # value = p.m if p.m < 1 else round(p.m,2) # FIXME: does not work for arrays
            unit = format(p.u, '~').replace(' ', '') # unit symbol without spaces
            stack.append((parameter.name, p.m, unit, parameter.is_default))
        
        df = pd.DataFrame(stack, columns='name value unit is_default'.split())
        if include_set_name:
            df.name = (self.name + '.') + df.name

        if name_include_unit:
            df.name = df.name + ' [' + df.unit + ']'

        return df

    def to_dict(self) -> Dict[str, Parameter]:
        """Return the dictionary representation of the instance."""
        return {parameter.name:parameter for parameter in self}

    def to_pickle(self, file:Path|None=None) -> Path:
        """Export the instance to a file in pickle format."""
        if not file:
            file = Path(self.name).with_suffix('.pickle')
        else:
            file = Path(file)
        
        with open(file, 'wb') as f:
            pickle.dump(self, f)
        
        return file

    def __iter__(self) -> Generator[Parameter]:
        """Iterate through the parameters"""
        for parm_name in self._parameters:
            yield getattr(self, parm_name)

    def __setattr__(self, attr, value):
        if self.LOCK_SETATTR and attr in self._parameters:
            raise SettingAttributeNotAllowed(f'It is not allowed to set {self.name}.{attr} directly. Use {self.name}.{attr}[<unit>] = {value} instead')
        object.__setattr__(self, attr, value)

    def copy(self) -> ParameterSet:
        """Return a deepcopy of self."""
        return deepcopy(self)
    
    
    def _to_tk(self, context, state='normal'):
        """Return as tk representation
        context: tk object - object in which to place the entries
        state: str -  can be disabled, normal, or readonly
        """
        frame = tk.LabelFrame(context, text=self.name, width=100, font=('Helvetica 9 bold'))
        
        entries = {}
        for row, parameter in enumerate(self):
            parameter:Parameter
            label, entry, unit = parameter._to_tk(frame)
            label.grid(row=row, column=0, padx=(2,2), pady=(2,2))
            entry.grid(row=row, column=1, padx=(2,2), pady=(2,2))
            unit.grid(row=row, column=2, padx=(2,2), pady=(2,2))
            entry.config(state=state)
            entries[parameter.name] = entry

        return frame, entries

    def __getitem__(self, name:str):
        dictionary = self.to_dict()
        return dictionary[name]

    def __hash__(self) -> hash:
        """Per default, the hash of a ParameterSet is a frozenset containing the string representation
        of the parameter names and their value.
        The unit is NOT included, hence the value should always be represented in the initial
        unit.
        """
        fz = []
        for name, parm in self.to_dict().items():
            fz.append(self.name + name + str(parm.normalized_value))
        return hash(frozenset(fz))

    def get_partial_hash(self, parameter_names:List[str]) -> hash:
        """Return the hash based on only the parameter names provided.
        """
        if not isinstance(parameter_names, (set, list, tuple)):
            raise TypeError(f"'parameter_names' must be a list, set or tuple. Got {type(parameter_names)}")
        stack = []
        for name in parameter_names:
            parm = self[name]
            string = self.name + name + str(parm.normalized_value)
            stack.append(string)
        return hash(frozenset(stack))



    
    
    



# Bokeh imports
from bokeh.plotting import figure, output_file, save
from bokeh.models import CustomJS, Select, OpenURL, TapTool, Div, ColumnDataSource, Button
from bokeh.models.tools import HoverTool
from bokeh.transform import linear_cmap
from bokeh.palettes import Spectral6, Spectral11, Turbo256

from bokeh.layouts import row, column
from bokeh.models import DataTable, DateFormatter, TableColumn, ColorBar, Circle
from bokeh.io import show as show_in_bokeh
from .abstracts import ParameterSet, Parameter, units
from typing import List, Type
import pandas as pd


class Analyze():
    """ParameterSets may be used to define input, but also the output from an certain analysis.
    If you have one or two such analyses, it may still be doable to go throught them numerically.
    Once we have more, it this might no longer be the case.

    
        ParameterSet1                                           ParameterSet4
        ParameterSet2   --input--> Calculation 1 --output-->    ParameterSet5
        ParameterSet3                                       


        ParameterSet1                                           ParameterSet4
        ParameterSet2   --input--> Calculation 2 --output-->    ParameterSet5
        ParameterSet3                                       

                                    ...etc




    
    """

    def __init__(self, sets:List[List[Type[ParameterSet]]]) -> None:
        self._check_sets(sets)
        self.sets = sets

    @staticmethod
    def _check_sets(sets:List[List[Type[ParameterSet]]]):
        """Check whether all input is correct"""
        for set in sets:
            for s in set:
                if not isinstance(s, ParameterSet):
                    raise TypeError('All items must be an instance of ParameterSet')

        aaa = []
        for set in sets:
            aaa.append({type(s) for s in set})
        
        for a in aaa[1:]:
            if not aaa[0] == a:
                raise TypeError('All sets of ParameterSets must be the same')
                
    
    def _sets_to_dataframe(self):
        """Convert all sets to 1 big dataframe"""
        sets = self.sets
        aa = []
        for ix, set in enumerate(sets):
            series = []
            for s in set:
                df = s.to_dataframe(name_include_unit=True) 
                df.set_index('name', inplace=True, drop=True)
                sr = df.value
                series.append(sr)
            ss = pd.concat(series)
            ss.name = str(ix)
            aa.append(ss)
        df = pd.concat(aa, axis=1).T
        return df

    def _get_single_scatter(self, source1:ColumnDataSource, source2:ColumnDataSource, show=False) -> column:
        """Return an interactive scatter for the current sets.
        
        Input:
            source1:ColumnDataSource


        Returns:
            A bokeh column[ row[Select, Select, Select]], figure]
        
        
        """
        ppp = figure(tools="pan,wheel_zoom,box_zoom,box_select,tap,undo,reset,save")
       
        ppp.toolbar.logo = None
        mapper = linear_cmap(field_name=f'ColorValues', palette=Turbo256 ,low=0 ,high=1)

        tags = source1.column_names
        tags.pop(tags.index('index'))
        tags.pop(tags.index('ColorValues'))
        x_start = tags[0]
        y_start = tags[1]

        scatter = ppp.circle(x=x_start, 
                             y=y_start,
                             color=mapper,
                             source=source1,
                             size=10,
                             fill_color=mapper, 
                             line_width=1)
        
        # The non-selected item should have constrasting color and a large alpha.
        scatter.nonselection_glyph = Circle(fill_color='pink', fill_alpha=0.75, line_color=None)

        ppp.xaxis.axis_label = x_start
        ppp.yaxis.axis_label = y_start
        
        
        ppp.toolbar.active_inspect = None
        
        # Sort of does want I want, but behaves erratic at times.
        js_link_selections = """ //link Cdf selection
        
        var s1 = source1
        var s2 = source2
        
        // impose the selected value to the other
        var ind1 = s1.selected.indices
        var ind2 = s2.selected.indices
        
        
        if (ind2.length == 0) {
            var ind = ind1
        } else {
            // intersection?
            var ind = ind1.filter(value => -1 !== ind2.indexOf(value))
        }
        
        //s1.selected.indices = ind
        s2.selected.indices = ind
        
        //source1.change.emit()
        source2.change.emit()
        
        """
        
        js = CustomJS(args=dict(source1=source1,
                                source2=source2,
                                ),
                      code=js_link_selections)
               
        source1.selected.js_on_change('indices', js)

        
        # url = "file://@PlatformID"
        # taptool = ppp.select(type=TapTool)
        # taptool.callback = OpenURL(url=url)
        
        hover = HoverTool()
        hover.tooltips = [(item,'@'+item) for item in tags]
        ppp.add_tools(hover)
        
        # Set what happens when you select item from the drop-down menu for X and Y axis data
        # NOTE: xaxis[0] has to be indexed, because there might be multiple axes
        xhandler = CustomJS(args=dict(axis=ppp.xaxis[0], 
                                      scatter=scatter),
                            code='''scatter.glyph.x = {field: cb_obj.value};
                                    axis.axis_label = cb_obj.value; 
                                 ''')
        
        
        xselect = Select(title="X-axis:", options=list(tags), value=x_start)
        xselect.js_on_change('value', xhandler)
        
        
        yhandler = CustomJS(args=dict(axis=ppp.yaxis[0], 
                                      scatter=scatter),
                            code='''scatter.glyph.y = {field: cb_obj.value};
                                    axis.axis_label = cb_obj.value; 
                                 ''')
        yselect = Select(title="Y-axis:", options=list(tags), value=y_start)
        yselect.js_on_change('value', yhandler)
        
        
        # Set what happens when you select item from the drop-down menu for color data
        # Color code # NOTE using fstring means using double brackets!  {{ }}
        codec = f"""
            
            var data = source1.data;
            var f = cb_obj.value;
            var x = data['ColorValues']
            
            //label.text = cb_obj.value.toString();
            
            var ind = source1.selected.indices
            
            
            // Update title
            ppp.title.text = 'Color: ' + cb_obj.value + ' - Selected: ' + ind.length
            
            if (cb_obj.value=="None") {{
                for (var i = 0; i < x.length; i++) {{
                    x[i] = 0
                }}
            }} else {{ 
            
                var y = data[f]
                var maxy = Math.max.apply(Math, y);
                var miny = Math.min.apply(Math, y);
                
                if (maxy==miny) {{
                    for (var i = 0; i < x.length; i++) {{
                       x[i] = 0
                    }}
                       
                }} else {{
                    for (var i = 0; i < x.length; i++) {{
                       x[i] = (y[i] - miny) / (maxy-miny)
                    }}
                }}
            }}
            
            source1.change.emit();
    
        """
        
        chandler = CustomJS(args=dict(source1=source1, ppp=ppp), code=codec)
        
        color_tags = ['None']
        color_tags.extend(list(tags))
        
        cselect = Select(title="Color:", options=color_tags, value='None')
        cselect.js_on_change('value', chandler)
        
        col = column(column(xselect, yselect, cselect), ppp)
        if show:
            show_in_bokeh(col)

        return col

    def get_double_scatter(self, show=False) -> row:
        """Return an interactive double scatter using Bokeh.
        A double scatter is a convenient way of navigating and filtering large parameteric
        data sets.
        """

        df = self._sets_to_dataframe()
        df = df.astype(float)
        df['ColorValues'] = 0.
        
        source1 = ColumnDataSource(df)
        source2 = ColumnDataSource(df)

        scat1 = self._get_single_scatter(source1, source2, show=False)
        scat2 = self._get_single_scatter(source2, source1, show=False)
        scatters = row(scat1, scat2)
        if show:
            show_in_bokeh(scatters)
        return scatters



        




from __future__ import annotations
from typing import List
from .abstracts import ParameterSet
import zipfile
from pathlib import Path
import tkinter as tk

class Container:
    """
    A collection of i/o parameters and any result files.

    This is an object-oriented datastorage, and propably pretty slow.
    But for more intensive stuff, we might have a relational database equivalent

    """

    i:List[ParameterSet]
    o:List[ParameterSet]

    def __init__(self, input:List[ParameterSet]=[], output:List[ParameterSet]=[]):
        self.input = input
        self.output = output
        self.additional_files = []
    
    def append_file(self, file:Path|str):
        self.additional_files.append(Path(file))

    def append_input(self, input:ParameterSet):
        self.input.append(input)
    
    def append_output(self, output:ParameterSet):
        self.output.append(output)

    @staticmethod
    def load(file:str, subclass:Container):
        input, output = [], []
        idict = {i.name:i for i in subclass.i}
        odict = {o.name:o for o in subclass.o}
        with zipfile.ZipFile(file, mode="r") as archive:
            for file in archive.filelist:
                data = archive.read(file.filename)
                name, _ = file.filename.split('.')
                
                if name in idict:
                    cls = idict[name]
                    inst = cls.from_serial_pickle(data)
                    input.append(inst)
                elif name in odict:
                    cls = odict[name]
                    inst = cls.from_serial_pickle(data)
                    output.append(inst)
                else:
                    pass #TODO
        return subclass(input=input, output=output)

    def save(self, file:Path):
        with zipfile.ZipFile(file, mode="w") as archive:
            for filename in self.additional_files:
                archive.write(filename)
            
            for pset in self.input + self.output:
                filename = pset.to_pickle()
                archive.write(filename)
            
    def file_ui(self):
        root = tk.Tk()
        
        input_column = tk.LabelFrame(root, text='INPUT', font=('Helvetica 10 bold'))
        sbi = tk.Scrollbar(input_column, orient = 'vertical')
        sbi.pack(side=tk.LEFT, fill=tk.Y)
        input_parm = tk.Frame(input_column, width=360, height=520)
        for inp in self.input:
            frame, entries = inp._to_tk(input_parm, state='disabled')
            frame.pack(padx=(10,10), pady=(10,10), side=tk.TOP)
        

        output_column = tk.LabelFrame(root, text='OUTPUT', font=('Helvetica 10 bold'))
        output_parm = tk.Frame(output_column, width=360, height=520)
        for inp in self.output:
            inp:ParameterSet
            frame, entries = inp._to_tk(output_parm, state='disabled')
            frame.pack(padx=(10,10), pady=(10,10), side=tk.TOP)
        
        input_parm.pack(side=tk.TOP)
        input_column.pack(side=tk.LEFT)
        output_parm.pack(side=tk.TOP)
        output_column.pack(side=tk.LEFT)
        

        root.mainloop()


    def get_partial_hash(self):
        pass
    
    def __hash__(self):
        pass

    def open(self):
        pass


    

class Vessel:
    """
    A collection of containers

    """

    def __init__(self, containers:List[Container]):
        self.containers = containers

    

class Database:
    """
    
    """
    pass


from .abstracts import ParameterSet, Parameter, units
import itertools
from typing import Iterable, List, Tuple, Dict

class Space:
    """Space facilitates constructing large parameter input spaces"""

    _expansions = []
    _criteria = []

    def __init__(self, parameter_set:ParameterSet) -> None:
        # if not issubclass(type(parameter_set), ParameterSet):
        #     raise TypeError(f'Expected ParameterSet, got {type(parameter_set)}')
        self.parameter_set = parameter_set

    def expand(self, parameter:str|Parameter, unit:str, space:Iterable[float|int]):
        if isinstance(parameter, Parameter):
            parameter = parameter.name
        self._expansions.append((parameter, unit, space))

    def add_criteria(self, parameter:str|Parameter, unit:str, f):
        if isinstance(parameter, Parameter):
            parameter = parameter.name
        self._criteria.append((parameter, unit, f))

    def construct(self) -> List[ParameterSet]:
        """Construct the space based on the expansions and criteria"""
        parameters = [exp[0] for exp in self._expansions]
        units = [exp[1] for exp in self._expansions]
        iterables = [exp[2] for exp in self._expansions]

        # Expand space
        space = []
        for values in itertools.product(*iterables):
            ps = self.parameter_set()
            for parameter, unit, value in zip(parameters, units, values):
                getattr(ps, parameter)[unit] = value
            space.append(ps)

        # Filter space
        if not self._criteria:
            return space
        else:
            filtered_space = []
            for crit, point in itertools.product(self._criteria, space):
                parameter, unit, criteria = crit
                value = getattr(point, parameter)[unit]
                if criteria(value):
                    filtered_space.append(point)
            return filtered_space

    def __len__(self):
        return len(self.construct())
    
        
    def __iter__(self):
        pass
        

    def evelope(self):
        """Get min and max values of the entire space"""
        raise NotImplementedError()

    def is_block_uniform(self):
        """Check whether the input space is block uniform, i.e. all Parameters are
        equally and uniformly represented. Block-uniform datasets are nice for data
        analysis because they are unbiased"""
        raise NotImplementedError()



if __name__ == "__main__":
    
    pass


import pytest
from pathlib import Path
from carg_io.abstracts import Parameter, ParameterSet, units, NaN
from carg_io.postprocessing import Analyze
from random import randint
import numpy as np
import itertools

__author__ = "eelco van Vliet"
__copyright__ = "eelco van Vliet"
__license__ = "MIT"


TEMP = Path(__file__).parent / 'temp'
TEMP.mkdir(exist_ok=True)


def test_post_process():
    

    class BoxI(ParameterSet):
        Length:Parameter = 1 * units.meter
        Width:Parameter = 1* units.meter
        Height:Parameter = 1* units.meter

    class BoxO(ParameterSet):
        Mass:Parameter = 1 * units.kg
        Volume:Parameter = 1 * units.meter**3
        EdgeLength:Parameter = 1 * units.meter
        SurfaceArea:Parameter = 1 * units.meter**2
        MomentOfInertiaY:Parameter = 1 * units.meter**4
        DeflectionPerkN:Parameter = 1 * units.meter

    


    def create_input_space():
        space = {
            "Length": np.linspace(2, 10, 10),
            "Width": np.linspace(2, 10, 10),
            "Height": np.linspace(2, 10, 10),
        }

        input = []
        for length, width, height in itertools.product(*space.values()):
            boxi = BoxI()
            boxi.Length['m'] = length
            boxi.Width['m'] = width
            boxi.Height['m'] = height
            input.append(boxi)
        return input

        
    def calculation(inputs):
        aaa = []
        for i in inputs:
            i:BoxI
            o = BoxO()
            l = i.Length['m']
            w = i.Width['m']
            h = i.Height['m']
            o.Volume['m**3'] = l * w * h
            o.EdgeLength['m'] = 4 * (l + w + h)
            o.Mass['kg'] = l * w * h * 7850
            o.SurfaceArea['m**2'] = 2*(l*h + h*w + w*l)
            I = o.MomentOfInertiaY['m**4'] = 1/12*w*h**3
            E = 210e9
            o.DeflectionPerkN['mm'] = 1e3*l/(48*E*I)*1e3
            aaa.append([i, o])

        return aaa
    
    design_sapce = create_input_space()
    result_space = calculation(design_sapce)

    analysis = Analyze(result_space)
    analysis.get_double_scatter(show=True)












